MODULE MMC;
(* =========================================================================  
   Astrobe Library Functions for MultiMedia / SD Cards using SPI
   
   REF: NXP Application Note AN10406. 
   "Accessing SD/MMC Card using SPI on LPC2000, V3 (Jan 3, 2007)" 
   
   Oberon Version (c) 2009-2012 CFB Software   
   http://www.astrobe.com  
   
   ========================================================================= *)

IMPORT SPI;
 
CONST  

(* SPI data is 8 bits. MMC uses 48 bits = 6 bytes *) 
  CmdSize = 6; 
 
(* The max MMC flash size is 256MB *) 
  BlockSize = 512;  (* 16-bit in size, 512 bytes *) 
  
  MaxTimeout = 0FFH; 
 
  NoError*            = 0;
  IdleStateTimeout*   = 1; 
  OpCondTimeout*      = 2; 
  SetBlockLenTimeout* = 3; 
  WriteBlockTimeout*  = 4; 
  WriteBlockFail*     = 5;
  ReadBlockTimeout*   = 6; 
  ReadBlockDataTokenMissing* = 7; 
  DataTokenTimeout*   = 8; 
  SelectCardTimeout*  = 9; 
  SetRelativeAddrTimeout* = 10; 
 
TYPE
  DataBlock = ARRAY BlockSize OF CHAR;
  CommandBlock = ARRAY CmdSize OF CHAR;


PROCEDURE Response(wanted: CHAR): BOOLEAN; 
VAR
  matched: BOOLEAN;
  count: INTEGER;
BEGIN 
  count := 0FFFH; 
  REPEAT 
    matched := (SPI.ReceiveChar() = wanted);
    DEC(count)
  UNTIL matched OR (count = 0);
  RETURN matched
END Response; 


PROCEDURE Init*(): INTEGER;
VAR
  i, status: INTEGER;
  data: DataBlock;
  cmd: CommandBlock; 
  dummy: CHAR;
BEGIN 
  status := NoError; 
 
  (* initialise the MMC card into SPI mode by sending 80 clks on *) 
  FOR i := 0 TO 9 DO data[i] := 0FFX END; 
  SPI.Send(data, 10); 
   
  (* send CMD0(RESET or GO_IDLE_STATE) command, all the arguments  
  are 00 FOR the reset command, precalculated checksum *) 
  cmd[0] := 40X; 
  cmd[1] := 00X; 
  cmd[2] := 00X; 
  cmd[3] := 00X; 
  cmd[4] := 00X; 
  cmd[5] := 95X; 
  SPI.Send(cmd, CmdSize); 
 
  IF ~Response(01X) THEN 
    status := IdleStateTimeout; 
  END; 
 
  IF status = NoError THEN 
    (* Send some dummy clocks after GO_IDLE_STATE *) 
    dummy := SPI.ReceiveChar(); 

    (* Keep sending command until zero response. *) 
    i := MaxTimeout; 
    REPEAT   
      (* send mmc CMD1(SEND_OP_COND) to bring out of idle state *) 
      (* all the arguments are 0x00 FOR command one *) 
      cmd[0] := 41X; 
      cmd[1] := 0X; 
      cmd[2] := 0X; 
      cmd[3] := 0X; 
      cmd[4] := 0X; 
      (* checksum is no longer required but we always send 0xFF *) 
      cmd[5] := 0FFX; 
      SPI.Send(cmd, CmdSize); 
      DEC(i) 
    UNTIL Response(0X) OR (i = 0); 
   
    (* timeout waiting FOR 0 from the MMC *) 
    IF i = 0 THEN 
      status := OpCondTimeout; 
    END 
  END;
  
  IF status = NoError THEN 
    (* Send some dummy clocks after SEND_OP_COND *) 
    dummy := SPI.ReceiveChar(); 
      
    (* send MMC CMD16(SET_BLOCKLEN) to set the block length *) 
    cmd[0] := 50X; 
    (* 00 00 02 00 set to 512 bytes *) 
    cmd[1] := 0X;   
    cmd[2] := 0X; 
    cmd[3] := 2X; 
    cmd[4] := 0X; 
    cmd[5] := 0FFX; 
    SPI.Send(cmd, CmdSize); 
     
    IF ~Response(0X) THEN 
      status := SetBlockLenTimeout; 
    END
  END; 

  IF status = NoError THEN  
    dummy := SPI.ReceiveChar()
  END; 
  
  RETURN status
END Init; 
 

PROCEDURE WriteFinished(): BOOLEAN; 
VAR
  count: INTEGER;
  finished: BOOLEAN;
BEGIN 
  count := 7FFFH; 
  REPEAT 
    finished := (SPI.ReceiveChar() # 0X);
    DEC(count)
  UNTIL finished OR (count = 0);
  RETURN finished
END WriteFinished; 


PROCEDURE SendCommand(command: CHAR; blockNumber: INTEGER);
VAR
  addr32, addrHigh, addrLow: INTEGER;
  cmd: CommandBlock;
BEGIN
  cmd[0] := command;
  addr32 := blockNumber * BlockSize;
  addrHigh := LSR(addr32, 16);
  addrLow := addr32 MOD 10000H;
  cmd[1] := CHR(LSR(addrHigh, 8)); 
  cmd[2] := CHR(addrHigh MOD 100H); 
  cmd[3] := CHR(LSR(addrLow, 8)); 
  cmd[4] := CHR(addrLow MOD 100H); 
  cmd[5] := 0FFX; 
  SPI.Send(cmd, CmdSize) 
END SendCommand;


PROCEDURE WriteBlock*(blockNumber: INTEGER; data: DataBlock): INTEGER;
VAR  
  status: INTEGER;
  dummy, statusCh: CHAR;
  checkSum: ARRAY 2 OF CHAR;
BEGIN 
  status := NoError;
   
  (* Send mmc CMD24(WRITE_SINGLE_BLOCK) to write the data to MMC card *) 
  SendCommand(58X, blockNumber);
   
  IF ~Response(0X) THEN 
    status := WriteBlockTimeout; 
  END; 
 
  IF status = NoError THEN 
    (* Set bit 0 to 0 to indicate the beginning of the data block *) 
    SPI.SendChar(0FEX); 
    SPI.Send(data, BlockSize); 
   
    (* Send dummy checksum *) 
    (* When the last checksum is sent, the response should come back 
    immediately. The status bits 3..0 should be 05 *)  
    checkSum[0] := 0FFX; 
    checkSum[1] := 0FFX; 
    SPI.Send(checkSum, 2); 
     
    statusCh := SPI.ReceiveChar(); 
    IF (ORD(statusCh) MOD 10H) # 05H THEN
      status := WriteBlockFail; 
    END
  END; 
 
  IF status = NoError THEN 
    IF ~WriteFinished() THEN 
      status := WriteBlockFail; 
    END
  END; 
 
  IF status = NoError THEN 
    dummy := SPI.ReceiveChar()
  END;
  RETURN status 
END WriteBlock; 
 

PROCEDURE ReadBlock*(blockNumber: INTEGER; VAR data: DataBlock): INTEGER; 
VAR
  status: INTEGER; 
  dummy: CHAR;
BEGIN 
  status := NoError;

  (* Send MMC CMD17(READ_SINGLE_BLOCK) to read the data from MMC card *) 
  SendCommand(51X, blockNumber);
 
  IF ~Response(0X) THEN 
    status := ReadBlockTimeout; 
  END; 
 
  (* Wait FOR data token *) 
  IF status = NoError THEN 
    IF ~Response(0FEX) THEN 
      status := ReadBlockDataTokenMissing; 
    END
  END; 
 
  IF status = NoError THEN 
    SPI.Receive(data, BlockSize); 
     
    (* Skip CRC bytes *) 
    dummy := SPI.ReceiveChar(); 
    dummy := SPI.ReceiveChar(); 
   
    dummy := SPI.ReceiveChar() 
  END;
  RETURN status 
END ReadBlock; 

END MMC.
